#!/usr/bin/python3
# -*- coding: utf-8 -*-
import os
import sys
import base64
import hashlib
import asyncio
import aiohttp
import async_timeout
import re
from tqdm import tqdm
import tempfile
import shutil
import numpy as np
import queue
import contextlib
from lxml import html
from cement.core import backend, foundation, controller, handler
from cement.ext.ext_argparse import ArgparseController, expose
from cement.core.foundation import CementApp
from cement.utils.misc import init_defaults

class AsyncTaskRunner(object):
    def __init__(self, max_job):
        self.tasks = asyncio.Queue()
        self.max_job = max_job
        self.futures = set()

    def put(self, task):
        self.tasks.put_nowait(task)

    def running_jobs(self):
        return len(self.futures)

    def empty(self):
        return self.tasks.empty()

    async def wait_for_some_job_done(self):
        done, self.futures = await asyncio.wait(self.futures, return_when = asyncio.FIRST_COMPLETED)
        return [await f for f in done]

    async def run(self):
        self.futures.add(asyncio.ensure_future(self.tasks.get()))
        while not self.empty() or self.running_jobs() != 1:
            for r in await self.wait_for_some_job_done():
                if r is not None:
                    while True:
                        if self.running_jobs() >= self.max_job:
                            await self.wait_for_some_job_done()
                        else:
                            break
                    self.futures.add(asyncio.ensure_future(r()))
                    self.futures.add(asyncio.ensure_future(self.tasks.get()))

    @classmethod
    @contextlib.contextmanager
    def with_default_with_runner(cls, max_job):
        AsyncTaskRunner.current_runner = AsyncTaskRunner(max_job)
        yield AsyncTaskRunner.current_runner


    @classmethod
    def job_creator(cls, func):
        def create(*args, **argv):
            r = cls.current_runner
            async def runner():
                await func(*args, **argv)
            r.put(runner)
        func.create = create
        return func

headers = {
        'Accept-Language':'zh-CN,zh;q=0.8',
        'Cookie' : 'watch_times=1'
}
bar_format = '{l_bar}{bar}|{n_fmt}/{total_fmt}[{elapsed}{rate_fmt}]'
chunk_size = 16 * 1024


def lastest_page(page):
    return "%s/v.php?next=watch&page=%d" % (config['base']['address'], page)


def category_page(category, page):
    return "%s/v.php?category=%s&viewtype=basic&page=%d" % (config['base']['address'], category, page)


def viewkey(id):
    return hashlib.md5(id.encode('utf-8')).hexdigest()[11:31]


def vid(id):
    return base64.standard_b64encode(id).decode('utf8')


class Video(object):
    def __init__(self, id, name, author, author_id):
        self.id = id
        self.name = name
        self.author = author
        self.author_id = author_id
        self.filename = '[%s]%s(%s)-%s.mp4' % (id, author, author_id, name)
        self.path = os.path.expanduser(os.path.join(config['base']['dir'], self.filename))

    def __str__(self):
        return self.filename

    def __repr__(self):
        return self.filename

    @AsyncTaskRunner.job_creator
    async def download(self, session):
        if os.path.exists(self.path) and config['base']['check_dup']:
            log.info("Video %s exists, skip." % (repr(self)))
            return True
        try:
            log.info('start downloading video %s...' % (repr(self)))
            metaurl = '%s/getfile_jw.php?VID=%s' % (config['base']['meta_address'], vid(self.id.encode('utf8')))
            async with session.get(metaurl) as response:
                url = re.search('file=(.*)$', await response.text()).group(1).replace('&', '?', 1)

            async with session.get(url) as response:
                size = int(response.headers.get('Content-Length'))
                if size < 200 * 1024:
                    log.debug('Content length is too short: %d' % size)
                    return False

            with tempfile.NamedTemporaryFile(delete= False) as temp:
                progress = tqdm(desc = self.id, unit="b", unit_scale=True, total=size)
                async def download_part(b, e):
                    retry_count = 0
                    while True:
                        try:
                            async with session.get(url, headers = {'Range': 'bytes=%d-%d' % (b, e)}) as response:
                                while True:
                                    chunk = await response.content.read(chunk_size)
                                    if not chunk or b >= e:
                                        return
                                    else:
                                        temp.seek(b)
                                        temp.write(chunk)
                                        b += len(chunk)
                                        progress.update(len(chunk))
                        except Exception as ex:
                            retry_count += 1
                            if retry_count % 5 == 0:
                                log.info("%s retry: %d" % (repr(self), retry_count))

                spacing = np.linspace(0, size, min(config['base']['slice'], max(1, size / (512 * 1024)))).astype(int)
                ranges = zip(spacing, spacing[1:])
                futures = set()
                await asyncio.wait([download_part(b, e-1) for b, e in ranges])
            shutil.move(temp.name, self.path)
            return True;
        except Exception as e:
            log.error("Failed to download video %s, error %s" % (repr(self), repr(e)))
            return False


@AsyncTaskRunner.job_creator
async def download_video_onpage(session, page):
    def to_video(elem):
        img = elem.xpath('(.//img[@title and @src])')[0]
        title = img.get('title')
        id = re.search('([0-9]+).jpg$', img.get('src')).group(1)
        author = elem.xpath('(.//a[@target="_parent"])')[0]
        author_id = re.search('UID=([0-9]+)$', author.get('href')).group(1)
        author = author.text
        return Video(id, title, author, author_id)

    async with session.get(page, headers = headers) as response:
        page_content = (await response.read()).decode('utf-8', 'ignore')
        dom = html.fromstring(page_content)
        videos = [to_video(v) for v in dom.xpath('//*[@class="listchannel"]') if v is not None]
    futures = []
    for v in videos:
        v.download.create(v, session)
        #await asyncio.sleep(1)


@AsyncTaskRunner.job_creator
async def download_video_by_viewkey(session, viewkey):
    try:
        url = '%s/view_video.php?viewkey=%s' % (config['base']['address'], viewkey)
        log.info("getting video info: %s..." % viewkey)
        async with session.get(url, headers = headers) as response:
            page_content = (await response.read()).decode('utf-8', 'ignore')
            dom = html.fromstring(page_content)
            title = dom.xpath('//*[@id="viewvideo-title"]')[0].text.strip()
            vid = re.search("VID=([0-9]+)&", page_content).group(1)
            author = re.search('receiver=(.*)"', page_content).group(1)
            author_id = re.search('UID=([0-9]+)"', page_content).group(1)
            video = Video(vid, title, author, author_id)
            log.info('video: %s: %s' % (viewkey, repr(video)))
        await video.download(session)
    except Exception as ex:
        log.error("Exception when downloading %s, error: %s" % (viewkey, repr(ex)))


@AsyncTaskRunner.job_creator
async def download_video_by_uid(session, uid):
    url = "%s/uvideos.php?UID=%s&type=public" % (config['base']['address'], uid)
    futures = []
    async with session.get(url, headers = headers) as response:
        log.info("getting pages for uid %s" % uid)
        pattern = re.compile(r'public&page=([0-9]+)"')
        page_content = (await response.read()).decode('utf-8', 'ignore')
        pages = range(1, max([1]+[int(page) for page in re.findall(pattern, page_content)]) + 1)
        log.info("Got %d pages for uid %s" % (len(pages), uid))
    for page in pages:
        url = "%s/uvideos.php?UID=%s&type=public&page=%d" % (config['base']['address'], uid, page)
        pattern = re.compile(r'viewkey=([0-9a-zA-Z]+)"')
        async with session.get(url, headers = headers) as response:
            page_content = (await response.read()).decode('utf-8', 'ignore')
            vks = set([m for m in re.findall(pattern, page_content)])
            for v in vks:
                download_video_by_viewkey.create(session, v)
                #await asyncio.sleep(1)

category = {
    'mf':'most favourite',
    'rf':'recent best',
    'hot': 'hot now',
    'rp': 'rate',
    'md': 'most discuessed',
    'tf':'most favourite this month',
    'top':'top this month'
}


def config_defaults():
    config = init_defaults('app')
    config['base'] = dict(
        address = "http://email.91dizhi.at.gmail.com.8h3.space",
        meta_address = "http://91.p9p.co",
        retry_count = 10,
        max_job = 20,
        slice = 10,
        check_dup = True
    )
    return config


class DownloadController(ArgparseController):
    class Meta:
        label = 'base'
        arguments = [
            (['--dir', '-d'], dict(help="directory to download videos", type=str)),
        ]

    @expose(
            arguments = [
                (['-c'], dict(help=', '.join([ '%s:%s' % (k,v) for k,v in category.items()]), type=str, choices=category.keys(), dest='category')),
                (['page'], dict(help='page to download', type=int, nargs='+')),
            ],
            help = 'download latest videos')
    def page(self):
        with aiohttp.ClientSession(loop=loop, connector = aiohttp.TCPConnector(limit = None)) as session:
            with AsyncTaskRunner.with_default_with_runner(20) as runner:
                if self.app.pargs.category is None:
                    [download_video_onpage.create(session, lastest_page(page)) for page in self.app.pargs.page]
                else:
                    [download_video_onpage.create(session, category_page(self.app.pargs.category, page)) for page in self.app.pargs.page]
                loop.run_until_complete(runner.run())

    @expose(
            arguments = [
                (['id'], dict(help='id to download, could be viewkey, id, url', nargs='+'))
            ],
            help = 'download video by id')

    def video(self):

        def guess_viewkey(s):
            if len(s) == 20:
                return s
            match = re.search('^([0-9]+)$', s)
            if match:
                return viewkey(match.group(1))
            match = re.search('viewkey=([0-9a-zA-Z]+)', s)
            if match:
                return match.group(1)
            self.app.log.error('Invalid id %s, skip' % s)
            return None

        vks = [guess_viewkey(vk) for vk in self.app.pargs.id if guess_viewkey(vk) is not None]
        with aiohttp.ClientSession(loop=loop, connector = aiohttp.TCPConnector(limit = None)) as session:
            with AsyncTaskRunner.with_default_with_runner(config['base']['max_job']) as runner:
                [download_video_by_viewkey.create(session, vk) for vk in vks]
                loop.run_until_complete(runner.run())

    @expose(
            arguments = [
                (['uid'], dict(help='id to download, could be viewkey, id, url'))
            ],
            help = 'download video by id')

    def uid(self):
        with aiohttp.ClientSession(loop=loop, connector = aiohttp.TCPConnector(limit = None, force_close=True)) as session:
            with AsyncTaskRunner.with_default_with_runner(config['base']['max_job']) as runner:
                download_video_by_uid.create(session, self.app.pargs.uid)
                loop.run_until_complete(runner.run())

def setup_global(app):
    global loop
    global log
    global config
    config = app.config
    loop = asyncio.get_event_loop()
    log = app.log

class App(CementApp):
    class Meta:
        label = '91downloader'
        extensions = ['colorlog']
        log_handler = 'colorlog'
        hooks =  [('post_argument_parsing', setup_global)]
        base_controller = DownloadController
        arguments_override_config = True
        config_defaults = config_defaults()
        config_files = [
            "~/.91.conf"
        ]

class DummyTqdmFile(object):
    """Dummy file-like that will write to tqdm"""
    file = None
    def __init__(self, file):
        self.file = file

    def write(self, x):
        # Avoid print() second call (useless \n)
        if len(x.rstrip()) > 0:
            tqdm.write(x, file=self.file)
    def isatty(self):
        return True
    def flush(self):
        return self.file.flush()

@contextlib.contextmanager
def stdout_redirect_to_tqdm():
    save_stdout = sys.stdout
    try:
        sys.stdout = DummyTqdmFile(sys.stdout)
        yield save_stdout
    except Exception as exc:
        raise exc
    finally:
        sys.stdout = save_stdout

if __name__ == '__main__':
    with stdout_redirect_to_tqdm() as save_stdout:
      with  App() as app:
          app.setup()
          app.run()
